package proxy

import (
	"crypto/tls"
	"crypto/x509"
	"errors"
	"fmt"
	"io/ioutil"
	"net/http"
	"net/http/httputil"
	"net/url"

	"github.com/gorilla/mux"

	"git.autistici.org/id/go-sso/httpsso"
)

// Backend defines a single-host HTTP proxy to a set of upstream
// backends.
type Backend struct {
	Host            string     `yaml:"host"`
	Upstream        []string   `yaml:"upstream"`
	ClientTLSConfig *TLSConfig `yaml:"client_tls"`
	//ServerTLSConfig *TLSConfig `yaml:"server_tls"`

	AllowedGroups []string `yaml:"allowed_groups"`
}

func (b *Backend) newHandler(ssow *httpsso.SSOWrapper) (http.Handler, error) {
	// Setup upstream connections.
	if len(b.Upstream) < 1 {
		return nil, errors.New("no backends specified")
	}

	u := &url.URL{Scheme: "http", Host: b.Host}
	if b.ClientTLSConfig != nil {
		u.Scheme = "https"
	}

	proxy := httputil.NewSingleHostReverseProxy(u)

	var tlsConfig *tls.Config
	if b.ClientTLSConfig != nil {
		var err error
		tlsConfig, err = b.ClientTLSConfig.toClientConfig()
		if err != nil {
			return nil, err
		}
	}
	proxy.Transport = newTransport(b.Upstream, tlsConfig)

	h := ssow.Wrap(proxy, b.Host+"/", b.AllowedGroups)
	return h, nil
}

// TLSConfig defines the TLS parameters for a client connection.
type TLSConfig struct {
	Cert string `yaml:"cert"`
	Key  string `yaml:"key"`
	CA   string `yaml:"ca"`
}

func (c *TLSConfig) toClientConfig() (*tls.Config, error) {
	cert, err := tls.LoadX509KeyPair(c.Cert, c.Key)
	if err != nil {
		return nil, err
	}

	cas, err := loadCA(c.CA)
	if err != nil {
		return nil, err
	}

	return &tls.Config{
		Certificates: []tls.Certificate{cert},
		RootCAs:      cas,
	}, nil
}

func loadCA(path string) (*x509.CertPool, error) {
	data, err := ioutil.ReadFile(path)
	if err != nil {
		return nil, err
	}
	cas := x509.NewCertPool()
	cas.AppendCertsFromPEM(data)
	return cas, nil
}

// func buildServerTLSConfig(config *Configuration) (*tls.Config, error) {
// 	var certs []tls.Certificate
// 	for _, b := range config.Backends {
// 		cert, err := tls.LoadX509KeyPair(b.ServerTLSConfig.Cert, b.ServerTLSConfig.Key)
// 		if err != nil {
// 			return nil, err
// 		}
// 		certs = append(certs, cert)
// 	}

// 	c := &tls.Config{
// 		Certificates: certs,
// 	}

// 	if config.CA != "" {
// 		cas, err := loadCA(config.CA)
// 		if err != nil {
// 			return nil, err
// 		}
// 		c.ClientAuth = tls.RequireAndVerifyClientCert
// 		c.ClientCAs = cas
// 	}

// 	c.BuildNameToCertificate()

// 	return c, nil
// }

// Configuration for the proxy.
type Configuration struct {
	SessionAuthKey string `yaml:"session_auth_key"`
	SessionEncKey  string `yaml:"session_enc_key"`
	//CA             string `yaml:"ca"`

	SSOLoginServerURL string `yaml:"sso_server_url"`
	SSOPublicKeyFile  string `yaml:"sso_public_key_file"`
	SSODomain         string `yaml:"sso_domain"`

	Backends []*Backend `yaml:"backends"`
}

// Sanity checks for the configuration.
func (c *Configuration) check() error {
	switch len(c.SessionAuthKey) {
	case 32, 64:
	case 0:
		return errors.New("session_auth_key is empty")
	default:
		return errors.New("session_auth_key must be a random string of 32 or 64 bytes")
	}
	switch len(c.SessionEncKey) {
	case 16, 24, 32:
	case 0:
		return errors.New("session_enc_key is empty")
	default:
		return errors.New("session_enc_key must be a random string of 16, 24 or 32 bytes")
	}
	if c.SSOLoginServerURL == "" {
		return errors.New("sso_server_url is empty")
	}
	if c.SSODomain == "" {
		return errors.New("sso_domain is empty")
	}
	return nil
}

// NewProxy builds a SSO-protected multi-host handler with the
// specified configuration.
func NewProxy(config *Configuration) (http.Handler, error) {
	if err := config.check(); err != nil {
		return nil, err
	}

	pkey, err := ioutil.ReadFile(config.SSOPublicKeyFile)
	if err != nil {
		return nil, err
	}

	w, err := httpsso.NewSSOWrapper(config.SSOLoginServerURL, pkey, config.SSODomain, []byte(config.SessionAuthKey), []byte(config.SessionEncKey))
	if err != nil {
		return nil, err
	}

	r := mux.NewRouter()
	for _, b := range config.Backends {
		h, err := b.newHandler(w)
		if err != nil {
			return nil, fmt.Errorf("error for host %s: %v", b.Host, err)
		}
		r.Host(b.Host).Handler(h)
	}
	return r, nil
}
